import copy
import io
import sys
import warnings
from typing import (
    List, Tuple, Set, Dict, Any,
    Optional, Union, Type, TypeVar
)
from frozendict import frozendict
from gelidum.exceptions import FrozenException
from gelidum.frozen import make_frozen_class, FrozenBase
from gelidum.utils import isbuiltin
from gelidum.typing import OnFreezeFuncType, OnUpdateFuncType

T = TypeVar('T')

_FrozenType = Optional[
    Union[
        bool, int, float, bytes, complex, str,
        bytes, frozendict, tuple, frozenset,
        Type[FrozenBase], T
    ]
]


def freeze(
        obj: T,
        on_update: Union[str, OnUpdateFuncType] = "exception",
        on_freeze: Union[str, OnFreezeFuncType] = "copy",
        inplace: Optional[bool] = None
        ) -> _FrozenType:

    # inplace argument will be removed from freeze in the next major version (0.5.0)
    if isinstance(inplace, bool):
        warnings.warn(
            DeprecationWarning(
                "Use of inplace is deprecated and will be removed in next major version (0.5.0)"
            )
        )
        if inplace:
            on_freeze_func: OnFreezeFuncType = __on_freeze_func(on_freeze="inplace")
        else:
            on_freeze_func: OnFreezeFuncType = __on_freeze_func(on_freeze="copy")

    else:
        on_freeze_func: OnFreezeFuncType = __on_freeze_func(on_freeze=on_freeze)

    on_update_func: OnUpdateFuncType = __on_update_func(on_update=on_update)

    return __freeze(obj=obj, on_update=on_update_func, on_freeze=on_freeze_func)


def __freeze(obj: Any, on_update: OnUpdateFuncType,
             on_freeze: OnFreezeFuncType) -> Any:

    if isbuiltin(obj):
        return obj

    class_name = type(obj).__name__
    freeze_func_name = f"__freeze_{class_name}"
    this_module = sys.modules[__name__]
    if hasattr(this_module, freeze_func_name):
        freeze_func = getattr(this_module, freeze_func_name)
        return freeze_func(obj, on_update=on_update, on_freeze=on_freeze)

    if isinstance(obj, object):
        return __freeze_object(obj, on_update=on_update, on_freeze=on_freeze)

    # Actually, this code is unreachable
    raise ValueError(f"object of type {obj.__class__} not frozen")  # pragma: no cover


def __freeze_bytearray(obj: bytearray, *args, **kwargs) -> bytes:
    return bytes(obj)


def __freeze_dict(obj: Dict, on_update: OnUpdateFuncType,
                  on_freeze: OnFreezeFuncType) -> frozendict:
    return frozendict({key: __freeze(value, on_update=on_update, on_freeze=on_freeze)
                       for key, value in obj.items()})


def __freeze_list(obj: List, on_update: OnUpdateFuncType,
                  on_freeze: OnFreezeFuncType) -> Tuple:
    return tuple(__freeze(item, on_update=on_update, on_freeze=on_freeze)
                 for item in obj)


def __freeze_tuple(obj: Tuple, on_update: OnUpdateFuncType,
                   on_freeze: OnFreezeFuncType) -> Tuple:
    return tuple(freeze(item, on_update=on_update, on_freeze=on_freeze)
                 for item in obj)


def __freeze_set(obj: Set, on_update: OnUpdateFuncType,
                 on_freeze: OnFreezeFuncType) -> frozenset:
    return frozenset([freeze(item, on_update=on_update, on_freeze=on_freeze)
                      for item in obj])


def __freeze_TextIOWrapper(*args, **kwargs) -> None:  # noqa
    raise io.UnsupportedOperation("Text file handlers can't be frozen")


def __freeze_BufferedWriter(*args, **kwargs) -> None:  # noqa
    raise io.UnsupportedOperation("Binary file handlers can't be frozen")


def __freeze_object(obj: object, on_update: OnUpdateFuncType,
                    on_freeze: OnFreezeFuncType) -> Type[FrozenBase]:

    frozen_obj = on_freeze(obj)
    for attr, value in frozen_obj.__dict__.items():
        attr_value = getattr(frozen_obj, attr)
        setattr(frozen_obj, attr, freeze(attr_value, on_update=on_update, on_freeze=on_freeze))
    frozen_obj.__class__ = make_frozen_class(
        klass=obj.__class__,
        attrs=list(obj.__dict__.keys()),
        on_update=on_update
    )
    return frozen_obj


def __on_freeze_func(on_freeze: Union[str, OnFreezeFuncType]) -> OnFreezeFuncType:
    if isinstance(on_freeze, str):
        if on_freeze == "copy":
            return lambda obj: copy.deepcopy(obj)
        elif on_freeze == "inplace":
            return lambda obj: obj
        else:
            raise AttributeError(
                f"Invalid value for on_freeze parameter, '{on_freeze}' found, "
                f"only 'copy' and 'inplace' are valid options if passed a string"
            )

    elif callable(on_freeze):
        return on_freeze

    else:
        raise AttributeError(
            f"Invalid value for on_freeze parameter, '{on_freeze}' found, "
            f"only 'copy', 'inplace' or a function are valid options"
        )


def __on_update_exception(frozen_obj: Type[FrozenBase], message: str, *args, **kwargs) -> None:
    raise FrozenException(message)


def __on_update_warning(frozen_obj: Type[FrozenBase], message: str, *args, **kwargs) -> None:
    warnings.warn(message)


def __on_update_func(on_update: OnUpdateFuncType) -> OnUpdateFuncType:
    if isinstance(on_update, str):
        if on_update == "exception":
            return __on_update_exception
        elif on_update == "warning":
            return __on_update_warning
        elif on_update == "nothing":
            return lambda message, *args, **kwargs: None
        else:
            raise AttributeError(
                f"Invalid value for on_update parameter, '{on_update}' found, "
                f"only 'exception', 'warning', and 'nothing' are valid options"
                f"if passed a string."
            )

    elif callable(on_update):
        return on_update

    else:
        raise AttributeError(
            f"Invalid value for on_update parameter, '{on_update}' found, "
            f"only 'exception', 'warning', 'nothing' or a function are "
            f"valid options"
        )


from gelidum.freeze import freeze  # noqa
from gelidum.decorators import freeze_params, freeze_final  # noqa
from gelidum.exceptions import FrozenException  # noqa
from gelidum.typing import Final  # noqa


import typing
from typing import Any, Callable, Type, TYPE_CHECKING
if TYPE_CHECKING:  # pragma: no cover
    from gelidum.frozen import FrozenBase  # noqa

try:
    _SpecialForm = getattr(typing, "_SpecialForm")

    @_SpecialForm
    def Final(self, parameters):  # noqa
        return typing.Final[parameters]

except AttributeError:  # pragma: no cover
    Final = typing.Final

_FrozenBase = Type["FrozenBase"]


OnUpdateFuncType = Callable[[_FrozenBase, str, ...], None]


OnFreezeFuncType = Callable[[Any], Any]



import sys
import threading
from typing import Type, List, cast, Dict
from gelidum.typing import OnUpdateFuncType


class FrozenBase(object):
    @classmethod
    def __gelidum_on_update(cls, *args, **kwargs):  # pragma: no cover
        raise NotImplementedError("Implement in derived class")

    @classmethod
    def get_gelidum_hot_class_name(cls) -> str:  # pragma: no cover
        raise NotImplementedError("Implement in derived class")

    @classmethod
    def get_gelidum_hot_class_module(cls) -> str:  # pragma: no cover
        raise NotImplementedError("Implement in derived class")

    def __setattr__(self, key, value):
        self.__gelidum_on_update(
            frozen_obj=self,
            message=f"Can't assign attribute '{key}' on immutable instance",
            key=key, value=value
        )

    def __set__(self, *args, **kwargs):
        self.__gelidum_on_update(
            frozen_obj=self,
            message="Can't assign setter on immutable instance",
            *args, **kwargs
        )

    def __delattr__(self, name):
        self.__gelidum_on_update(
            frozen_obj=self,
            message=f"Can't delete attribute '{name}' on immutable instance",
            name=name
        )

    def __setitem__(self, key, value):
        self.__gelidum_on_update(
            frozen_obj=self,
            message=f"Can't set key '{key}' on immutable instance",
            key=key, value=value
        )

    def __delitem__(self, key):
        self.__gelidum_on_update(
            frozen_obj=self,
            message=f"Can't delete key '{key}' on immutable instance",
            key=key)

    def __reversed__(self):
        self.__gelidum_on_update(
            frozen_obj=self,
            message="Can't reverse on immutable instance"
        )


__FROZEN_CLASSES: Dict[str, Type[FrozenBase]] = dict()
__FROZEN_CLASSES_LOCK = threading.Lock()


def __store_frozen_class(
        klass: Type[object], frozen_class: Type[FrozenBase]
) -> None:
    """
    Add a frozen class to this module.
    Required for pickle serialization as only objects of non-dynamic
    classes are allowed.
    This method is thread-safe.
    :param frozen_class: a class that inherits from FrozenBase.
    """
    with __FROZEN_CLASSES_LOCK:
        klass_key = f"{klass.__module__}.{klass.__qualname__}"
        __FROZEN_CLASSES[klass_key] = frozen_class
        # Required for pickling frozen objects (only classes defined in actual
        # modules can have their objects pickled)
        setattr(sys.modules[__name__], frozen_class.__name__, frozen_class)


def clear_frozen_classes() -> None:
    with __FROZEN_CLASSES_LOCK:
        __FROZEN_CLASSES.clear()


def __create_frozen_class(
        klass: Type[object],
        attrs: List[str],
        on_update_func: OnUpdateFuncType
) -> Type[FrozenBase]:
    camel_case_module = (
        klass.__module__.title().replace(".", "").replace("_", "")
    )
    frozen_class_name = f"Frozen{klass.__name__}From{camel_case_module}"
    frozen_class: Type[FrozenBase] = cast(
        Type[FrozenBase],
        type(
            frozen_class_name,
            (klass, FrozenBase),
            {
                "__slots__": tuple(),
                **{
                    'get_gelidum_hot_class_name': lambda _: klass.__name__,
                    'get_gelidum_hot_class_module': lambda _: klass.__module__,
                    '_FrozenBase__gelidum_on_update':
                        lambda _self, *args, **kwargs:
                        on_update_func(*args, **kwargs),
                    **{attr: None for attr in attrs}
                }
            }
        )
    )
    __store_frozen_class(klass=klass, frozen_class=frozen_class)
    return frozen_class


def make_frozen_class(klass: Type[object], attrs: List[str],
                      on_update: OnUpdateFuncType) -> Type[FrozenBase]:
    klass_key = f"{klass.__module__}.{klass.__qualname__}"
    with __FROZEN_CLASSES_LOCK:
        frozen_class = __FROZEN_CLASSES.get(klass_key)

    if not frozen_class:
        frozen_class = __create_frozen_class(
            klass=klass, attrs=attrs,
            on_update_func=on_update
        )

    return frozen_class



from __future__ import annotations

import functools
from typing import Optional, Iterable, Set
from gelidum import freeze


def freeze_params(params: Optional[Iterable[str]] = None):
    def inner_freeze_params(func):
        """Freeze all input params of a method"""
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            func_args = tuple([freeze(arg, on_freeze="copy")
                               for arg in args])
            func_kwargs = {
                kwarg_name: freeze(kwarg, on_freeze="copy")
                if kwarg_name in params else kwarg
                for kwarg_name, kwarg in kwargs.items()
            }
            return func(*func_args, **func_kwargs)
        return wrapper
    return inner_freeze_params


def freeze_final(func):
    """Freeze all final params of a method"""
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        unnamed_params_to_freeze: Set[str] = {
            i
            for i, (param_name, param_typing) in
            enumerate(func.__annotations__.items())
            if str(param_typing).startswith("typing.Final")
        }
        named_params_to_freeze: Set[str] = {
            param_name
            for param_name, param_typing in
            func.__annotations__.items()
            if str(param_typing).startswith("typing.Final")
        }
        func_args = tuple([
            (
                freeze(arg, on_freeze="copy")
                if arg_index in unnamed_params_to_freeze else
                arg
            )
            for arg_index, arg in enumerate(args)
            ])
        func_kwargs = {
            kwarg_name: (
                freeze(kwarg, on_freeze="copy")
                if kwarg_name in named_params_to_freeze else
                kwarg
            )
            for kwarg_name, kwarg in kwargs.items()
        }
        return func(*func_args, **func_kwargs)
    return wrapper
